@@ -6,14 +6,12 @@ try:  # pragma: no cover
     import aesara
     import aesara.tensor as at
     from aesara.graph.op import get_test_value
-    from aesara.graph.utils import TestValueError
     from aesara.scalar import upcast
     from aesara.tensor.extra_ops import broadcast_to as at_broadcast_to
 except ImportError:  # pragma: no cover
     import theano as aesara
     import theano.tensor as at
     from theano.graph.op import get_test_value
-    from theano.graph.utils import TestValueError
     from theano.scalar import upcast
     from theano.tensor.extra_ops import broadcast_to as at_broadcast_to
 
@@ -29,7 +27,7 @@ from pymc3.distributions.mixture import _conversion_map, all_discrete
 from pymc3_hmm.utils import tt_broadcast_arrays, vsearchsorted
 
 
-def distribution_subset_args(dist, shape, idx, point=None):
+def distribution_subset_args(dist, shape, idx):
     """Obtain subsets of a distribution parameters via indexing.
 
     This is used to effectively "lift" slices/`Subtensor` `Op`s up to a
@@ -52,45 +50,21 @@ def distribution_subset_args(dist, shape, idx, point=None):
         to (naively) determine the parameters' broadcasting pattern.
     idx : ndarray or TensorVariable
         The indices applied to the parameters of `dist`.
-    point : dict (optional)
-        A dictionary keyed on the `str` names of each parameter in `dist`,
-        which are mapped to NumPy values for the corresponding parameter.  When
-        this is given, the Theano parameters are replaced by their values in the
-        dictionary.
 
     Returns
     -------
-    res: list
-        An ordered set of broadcasted and indexed parameters for `dist`.
-
+    A ``dict`` of the broadcasted and indexed parameters for `dist`.
 
     """
 
     dist_param_names = dist._distr_parameters_for_repr()
 
-    if point:
-        # Try to get a concrete/NumPy value if a `point` parameter was
-        # given.
-        try:
-            idx = get_test_value(idx)
-        except TestValueError:  # pragma: no cover
-            pass
-
-    res = []
+    res = dict()
     for param in dist_param_names:
 
-        # Use the (sampled) point, if present
-        if point is None or param not in point:
-            x = getattr(dist, param, None)
+        bcast_res = at_broadcast_to(getattr(dist, param), shape)
 
-            if x is None:
-                continue
-        else:
-            x = point[param]
-
-        bcast_res = at_broadcast_to(x, shape)
-
-        res.append(bcast_res[idx])
+        res[param] = bcast_res[idx]
 
     return res
 
@@ -194,7 +168,8 @@ class SwitchingProcess(Distribution):
         for i, dist in enumerate(self.comp_dists):
             i_mask = at.eq(self.states, i)
             obs_i = obs_tt[i_mask]
-            subset_dist = dist.dist(*distribution_subset_args(dist, obs.shape, i_mask))
+            subset_dist_kwargs = distribution_subset_args(dist, obs.shape, i_mask)
+            subset_dist = dist.dist(**subset_dist_kwargs)
             logp_val = at.set_subtensor(logp_val[i_mask], subset_dist.logp(obs_i))
 
         return logp_val
@@ -251,10 +226,10 @@ class SwitchingProcess(Distribution):
             i_idx = np.where(expanded_states == i)
             i_size = len(i_idx[0])
             if i_size > 0:
-                subset_args = distribution_subset_args(
-                    dist, expanded_states.shape, i_idx, point=point
+                subset_kwargs = distribution_subset_args(
+                    dist, expanded_states.shape, i_idx
                 )
-                state_dist = dist.dist(*subset_args)
+                state_dist = dist.dist(**subset_kwargs)
 
                 sample = state_dist.random(point=point)
                 samples[i_idx] = sample
@@ -477,3 +452,34 @@ class Constant(Distribution):
 
     def _distr_parameters_for_repr(self):
         return ["c"]
+
+
+class HorseShoe(pm.Normal):
+    """
+    A Horseshoe distribution
+
+    """
+
+    def __init__(self, tau, shape, **kwargs):
+        self.tau = tau
+        self.beta_raw = pm.Normal.dist(tau=1, shape=shape)
+        self.lmbda = pm.HalfCauchy.dist(beta=1, shape=shape)
+        super().__init__(shape=shape, **kwargs)
+
+    def random(self, point=None, size=None):
+        beta_raw = self.beta_raw.random(point=point, size=size)
+        lmbda = self.lmbda.random(point=point, size=size)
+        return beta_raw * lmbda * self.tau
+
+    def logp(self, values):
+        """
+        XXX: This is only a placeholder for the log-probability.
+        It is required to get past PyMC3 design restrictions.
+        Do not use this distribution without the `HSStep` sampler!
+
+        """
+        warnings.warn(
+            "logp of HorseShoe class is not implemented."
+            "Do not use this distribution without the `HSStep` sampler!"
+        )
+        return 0
