@@ -1,16 +1,20 @@
 from itertools import chain
+from typing import Callable, Tuple
 
 import numpy as np
 
 try:  # pragma: no cover
     import aesara.scalar as aes
     import aesara.tensor as at
+    from aesara import config
     from aesara.compile import optdb
     from aesara.graph.basic import Variable, graph_inputs
     from aesara.graph.fg import FunctionGraph
     from aesara.graph.op import get_test_value as test_value
     from aesara.graph.opt import OpRemove, pre_greedy_local_optimizer
     from aesara.graph.optdb import Query
+    from aesara.scalar.basic import Dot
+    from aesara.sparse.basic import StructuredDot
     from aesara.tensor.elemwise import DimShuffle, Elemwise
     from aesara.tensor.subtensor import AdvancedIncSubtensor1
     from aesara.tensor.var import TensorConstant
@@ -26,13 +30,20 @@ except ImportError:  # pragma: no cover
     from theano.tensor.elemwise import DimShuffle, Elemwise
     from theano.tensor.subtensor import AdvancedIncSubtensor1
     from theano.tensor.var import TensorConstant
+    from theano.tensor.basic import Dot
+    from theano.sparse.basic import StructuredDot
+    from theano import config
+
+from functools import singledispatch
 
 import pymc3 as pm
+import scipy
 from pymc3.distributions.distribution import draw_values
 from pymc3.step_methods.arraystep import ArrayStep, BlockedStep, Competence
 from pymc3.util import get_untransformed_name
+from scipy.stats import invgamma
 
-from pymc3_hmm.distributions import DiscreteMarkovChain, SwitchingProcess
+from pymc3_hmm.distributions import DiscreteMarkovChain, HorseShoe, SwitchingProcess
 from pymc3_hmm.utils import compute_trans_freqs
 
 big: float = 1e20
@@ -452,3 +463,229 @@ class TransMatConjugateStep(ArrayStep):
             return Competence.COMPATIBLE
 
         return Competence.INCOMPATIBLE
+
+
+def large_p_mvnormal_sampler(D_diag, Phi, a):
+    r"""Efficiently sample from a large multivariate normal.
+
+    This function draws samples from the following distribution:
+
+    .. math::
+       \beta \sim \operatorname{N}\left( \mu, \Sigma \right)
+
+    where
+
+    .. math::
+       \mu = \Sigma \Phi^\top a, \\
+       \Sigma = \left( \Phi^\top \Phi + D^{-1} \right)^{-1}
+
+    and :math:`a \in \mathbb{R}^{n}`, :math:`\Phi \in \mathbb{R}^{n \times p}`.
+
+    This approach is particularly effective when :math:`p \gg n`.
+
+    From "Fast sampling with Gaussian scale-mixture priors in high-dimensional
+    regression", Bhattacharya, Chakraborty, and Mallick, 2015.
+
+    """
+    N = a.shape[0]
+    u = np.random.normal(0, np.sqrt(D_diag))
+    delta = np.random.normal(size=N)
+    if scipy.sparse.issparse(Phi):
+        Phi_D = Phi.multiply(D_diag)
+        v = Phi * u + delta
+        Z = (Phi_D * Phi.T + scipy.sparse.eye(N)).toarray()
+        w = scipy.linalg.solve(Z, a - v, assume_a="sym")
+        beta = u + Phi_D.T * w
+    else:
+        Phi_D = Phi * D_diag
+        v = Phi.dot(u) + delta
+        Z = Phi_D.dot(Phi.T)
+        Z.flat[:: N + 1] += 1
+        w = scipy.linalg.solve(Z, a - v, assume_a="sym")
+        beta = u + Phi_D.T @ w
+    return beta
+
+
+def hs_step(
+    lambda2: np.ndarray,
+    tau2: np.ndarray,
+    vi: np.ndarray,
+    xi: np.ndarray,
+    X: np.ndarray,
+    y: np.ndarray,
+):
+    _, M = X.shape
+
+    D_diag = tau2 * lambda2
+    beta = large_p_mvnormal_sampler(D_diag, X, y)
+    beta2 = beta ** 2
+
+    lambda2 = invgamma(a=1, scale=1 / vi + beta2 / (2 * tau2)).rvs()
+    tau2 = invgamma(a=(M + 1) / 2, scale=1 / xi + (beta2 / lambda2).sum() / 2).rvs()
+    vi = invgamma(a=1, scale=1 + 1 / lambda2).rvs()
+    xi = invgamma(a=1, scale=1 + 1 / tau2).rvs()
+
+    return beta, lambda2, tau2, vi, xi
+
+
+@singledispatch
+def hs_regression_model(dist: pm.Distribution, rv, model) -> Tuple[Callable, Variable]:
+    """Determine the normal regression model for a Horseshoe sampler.
+
+    Return a function that computes the normal regression: i.e. the observation
+    vector and regression matrix.
+
+    For non-normal distributions, the normal regression model is an
+    approximation (e.g. Polya-Gamma).
+    """  # noqa: E501
+    raise NotImplementedError()
+
+
+@hs_regression_model.register(pm.Normal)
+def hs_regression_model_Normal(dist, rv, model):
+    mu = dist.mu
+    y_X_fn = None
+    if hasattr(rv, "observations"):
+
+        def y_X_fn(points, X):
+            return rv.observations, X
+
+    return y_X_fn, mu
+
+
+@hs_regression_model.register(pm.NegativeBinomial)
+def hs_regression_model_NegativeBinomial(dist, rv, model):
+
+    mu = at.as_tensor_variable(dist.mu)
+
+    if mu.owner and mu.owner.op == at.exp:
+        eta = mu.owner.inputs[0]
+    else:
+        eta = mu
+
+    alpha = at.as_tensor_variable(dist.alpha)
+    if hasattr(rv, "observations"):
+        from polyagamma import random_polyagamma
+
+        # pm.model.fn.signiturw
+        h_z_alpha_fn = model.fn(
+            [alpha + rv.observations, eta.squeeze() - at.log(alpha), alpha]
+        )
+
+        def y_X_fn(points, X):
+            h, z, alpha = h_z_alpha_fn(points)
+
+            omega = random_polyagamma(h, z)
+
+            V_diag_inv = np.abs(omega)
+            sigma2 = 1 / V_diag_inv
+            sigma = np.sqrt(sigma2)
+
+            if scipy.sparse.issparse(X):
+                Phi = (X.T.multiply(np.sqrt(V_diag_inv))).T
+            else:
+                Phi = (X.T * np.sqrt(V_diag_inv)).T
+
+            y_aug = np.log(alpha) + (rv.observations - alpha) / (2.0 * omega)
+            y_aug = (y_aug / sigma).astype(config.floatX)
+            return y_aug, Phi
+
+        return y_X_fn, eta
+
+    return None, eta
+
+
+class HSStep(BlockedStep):
+    name = "hsgibbs"
+
+    def __init__(self, vars, values=None, model=None):
+        model = pm.modelcontext(model)
+
+        if len(vars) > 1:
+            raise ValueError("This sampler only takes one variable.")
+
+        (beta,) = pm.inputvars(vars)
+
+        if not isinstance(beta.distribution, HorseShoe):
+            raise TypeError("This sampler only samples `HorseShoe`s.")
+
+        other_model_vars = [
+            value for attr, value in model.named_vars.items() if value != beta
+        ]
+        y_X_fn, X_fn = None, None
+
+        for var in other_model_vars:
+            # Look through all the attributes of the variable and see if any of
+            # the parameters have a multiplication relationship with the
+            # Horseshoe variable
+            if hasattr(var, "distribution"):
+                try:
+                    y_X_fn, eta = hs_regression_model(var.distribution, var, model)
+                except NotImplementedError:
+                    continue
+            elif isinstance(var, pm.model.DeterministicWrapper):
+                eta = var.owner.inputs[0]
+
+            dense_dot = eta.owner and isinstance(eta.owner.op, Dot)
+            sparse_dot = eta.owner and isinstance(eta.owner.op, StructuredDot)
+
+            dense_inputs = dense_dot and beta in eta.owner.inputs
+            sparse_inputs = sparse_dot and beta in eta.owner.inputs[1].owner.inputs
+
+            if not (dense_inputs or sparse_inputs):
+                continue
+
+            if not y_X_fn:
+                # We don't have the observation distribution, so we need to
+                # find it.  This happens when a `Deterministic` bridges a
+                # `Horseshoe` parameter with it's observation distribution's
+                # mean.
+                y_X_fn = None
+                obs_mu = None
+                for obs_rv in model.observed_RVs:
+                    try:
+                        y_X_fn, obs_mu = hs_regression_model(
+                            obs_rv.distribution, obs_rv, model
+                        )
+                        break
+                    except NotImplementedError:
+                        continue
+
+                # The `Deterministic` should be the mean parameter of the
+                # observed distribution
+                if var != obs_mu:
+                    continue
+
+            if dense_inputs:
+                X_fn = model.fn(eta.owner.inputs[1].T)
+            else:
+                X_fn = model.fn(eta.owner.inputs[0])
+
+        if not (X_fn and y_X_fn):
+            raise NotImplementedError(
+                f"Cannot find a design matrix or dependent variable associated with {beta}"  # noqa: E501
+            )
+
+        self.vars = [beta]
+
+        M = model.test_point[beta.name].shape[-1]
+
+        # if observation dist is normal then y_aug_fn = y_fn when it is NB
+        # then, hs_regression_model, dispatch i.distribution...
+
+        self.vi = np.full(M, 1)
+        self.lambda2 = np.full(M, 1)
+        self.beta = np.full(M, 1)
+        self.tau2 = 1
+        self.xi = 1
+        self.y_X_fn = y_X_fn
+        self.X_fn = X_fn
+
+    def step(self, point):
+        X = self.X_fn(point)
+        y, X = self.y_X_fn(point, X)
+        self.beta, self.lambda2, self.tau2, self.vi, self.xi = hs_step(
+            self.lambda2, self.tau2, self.vi, self.xi, X, y
+        )
+        point[self.vars[0].name] = self.beta
+        return point
